#!/bin/python3

# Watchman © 2023

print('''



 \ \        /       |          |                              
  \ \  \   /  _` |  __|   __|  __ \   __ `__ \    _` |  __ \  
   \ \  \ /  (   |  |    (     | | |  |   |   |  (   |  |   | 
    \_/\_/  \__,_| \__| \___| _| |_| _|  _|  _| \__,_| _|  _| 
                                                              
        Forensic Extractor for YouTube - version 1.0.0


                        thewatchman.pro
                            © 2023







''')


import sys
import logging
from colr import color
import csv
import datetime
import hashlib
import socket
import os
import math
import logging
import shutil
import time
import subprocess
import re
import yt_dlp
import http.client
import urllib.request
from datetime import datetime
from blake3 import blake3
from pytube import YouTube, exceptions
from pytube.exceptions import VideoUnavailable
from pytube.cli import on_progress
from yt_dlp import YoutubeDL



# Prompt user for a CSV file containing URLs and parse them.
def read_urls_from_csv():

    for attempts in range(5):
        try:
            csv_path = input("Please enter the file path for the CSV file:\n")

            # Validation 1: Verify that path exists.
            if not os.path.isfile(csv_path):
                logging.info(color('\nInvalid file path detected.\n', fore="f26400"))
                continue

            csv_extension = os.path.splitext(csv_path)[1]

            if csv_extension == ".csv" or csv_extension == ".txt":
                urls = []
                with open(csv_path, 'r', encoding='utf-8') as csv_file:
                    reader = csv.reader(csv_file)
                    for row in reader:
                        for url in row:

                            # Validation 2: Strip empty spaces from URL.
                            url = url.strip()

                            if url != "":

                                # Validation 3: Remove duplicate URL entries.
                                if url not in urls:
                                    urls.append(url)
                            else:
                                continue
                return urls
            
            else:
                logging.info(color('\nInvalid file extension detected: \n', fore="f26400") + f"{csv_extension}\n")
        
        except Exception as e:
            logging.info(color(f"\nAn undefined error occurred:", fore="f26400"))
            logging.info(f"{e}\n")

        except KeyboardInterrupt:
            logging.info("\nTermination request received. Exiting...\n")
            sys.exit()

    logging.info(color('\nNo valid file path located. Exiting...\n\n', fore="f26400"))
    sys.exit()


# Perform sanitization on URLs to prevent malicious entries.
def sanitize_urls(urls):
    
    try:
        whitelist = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789/_-'
        url_pattern = re.compile(r'^https?://(?:www\.)?(?:youtu\.be/|youtube\.com/watch\?v=)([a-zA-Z0-9_-]+[a-zA-Z0-9_-]*)$')
        valid_urls = []
        sanitized_urls = []

        for url in urls:
            url_match = url_pattern.match(url)
            
            if not url_match:
                logging.info(color('\nInvalid URL detected: ', fore="f26400") + url)
                continue
            
            else:
                valid_urls.append(url)

        for url in valid_urls:
            for char in url:
                if char not in whitelist:
                    break
                else:
                    continue

            sanitized_urls.append(url)
        return sanitized_urls
        

    except Exception as e:
        logging.info(color('An error has occurred.\n', fore="f26400"))
        logging.info(e)
        logging.info(url)

    except KeyboardInterrupt:
        logging.info("\nTermination request received. Exiting...\n")
        sys.exit()



# Verify whether CSV file exists and write column headers to it.
def check_csv():

    try:
        csv_columns = ['File Name (Video)', 'File Path (Video)', 'URL', 'Title', 'View Count', 'Like Count', 'Comment Count', 'Location', 'Duration', 'Channel', 'Extraction Time (Video)', 'Release Time (Video)', 'Upload Time', 'BLAKE3 (Video)', 'SHA-512 (Video)', 'SHA-256 (Video)', 'SHA-1 (Video)', 'MD5 (Video)', 'File Name (Audio)', 'File Path (Audio)', 'Extraction Time (Audio)', 'BLAKE3 (Audio)', 'SHA-512 (Audio)', 'SHA-256 (Audio)', 'SHA-1 (Audio)', 'MD5 (Audio)', 'File Name (Multiplex)', 'File Path (Multiplex)', 'SHA-512 (Multiplex)', 'Multiplexing Start Time', 'Multiplexing End Time']
        csv_log = "extraction_log.csv"
        
        current_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
        logging.info(color(f"\n{current_time}", fore="2d75ce"))
        
        if os.path.exists(csv_log):
            with open(csv_log, 'r', encoding='utf-8') as log_file:
                reader = csv.reader(log_file)
                existing_columns = next(reader, [])
                if existing_columns == csv_columns:
                    logging.info("Column headers detected. Proceeding...\n")
                else:
                    logging.info("Column headers created. Proceeding...\n")
                    with open(csv_log, 'w', encoding='utf-8', newline='') as log_file:
                        writer = csv.writer(log_file)
                        writer.writerow(csv_columns)
        else:
            logging.info("CSV file created in the current directory. Proceeding...\n")
            with open(csv_log, 'w', encoding='utf-8', newline='') as log_file:
                writer = csv.writer(log_file)
                writer.writerow(csv_columns)
        return csv_log
    
    except Exception as e:
        logging.info(color(f"\nAn undefined error occurred:", fore="f26400"))
        logging.info(f"{e}\n")

    except KeyboardInterrupt:
        logging.info("\nTermination request received. Exiting...\n")
        sys.exit()



# Retrieve YouTube videos using the URL list.
def download_data(sanitized_urls, csv_log):
    
    invalid_urls = []
    max_retries = 70
    network_retry = 5

    try:
        with open(csv_log, 'a', encoding='utf-8', newline='') as log_file:
            writer = csv.writer(log_file)
            number = 0

            for url in sanitized_urls:
                number += 1
                retry_count = 0
                network_error = 0
                sleep = 10
                
                try:
                    video_initiation_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                    logging.info("\n-------------------------      PHASE 1      -------------------------\n")
                    logging.info(color(f"\n{video_initiation_time}", fore="2d75ce"))
                    logging.info(f"Initiating extraction process from {url}. Please wait...\n")

                    yt = YouTube(url)
                    player_config_args = yt._player_config_args
                    
                    video_id = re.search(r"v=([\w-]+)", url).group(1)
                        
                except AttributeError:
                    try:
                        video_id = re.search(r"be/([\w-]+)", url).group(1)
                    
                    except AttributeError:
                        logging.info(f"Unable to process {url}.")
                        break
                
                except ValueError:
                    logging.info(color('Invalid URL detected. Proceeding...\n', fore="f26400"))
                    time.sleep(5)
                    continue
                
                except Exception as e:
                    logging.info("The following error occurred:")
                    logging.info(e)
                    break
                    
                except KeyboardInterrupt:
                    logging.info("\nTermination request received. Exiting...\n")
                    sys.exit()


                #Retrieve metadata using YouTube DLP        
                try:
                    options = {
                        'format': 'bestvideo[ext=mp4]',
                        'quiet': True,
                        'no-warnings': True,
                        'ignoreerrors': True
                    }

                    with yt_dlp.YoutubeDL(options) as ydl:
                        metadata_start_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                        logging.info("\n-------------------------      PHASE 2      -------------------------\n")
                        logging.info(color(f"\n{metadata_start_time}", fore="2d75ce"))
                        logging.info(f"Extracting metadata from {url}.\n")
                        ydl_metadata = ydl.extract_info(url, download=False)
                    
                    video_title = ydl_metadata.get('title')
                    location = ydl_metadata.get('location')
                    view_count = ydl_metadata.get('view_count')
                    like_count = ydl_metadata.get('like_count')
                    comment_count = ydl_metadata.get('comment_count')
                    duration = ydl_metadata.get('duration_string')
                    channel = ydl_metadata.get('channel')
                    upload_date = ydl_metadata.get('upload_date')
                    release_time_unix = ydl_metadata.get('release_timestamp')

                    upload_year = upload_date[:4]
                    upload_month = upload_date[4:6]
                    upload_day = upload_date[6:]
                    upload_date_readable = f"{upload_year}-{upload_month}-{upload_day}"


                    if release_time_unix is not None:
                        release_time_readable = (datetime.fromtimestamp(release_time_unix)).strftime("%Y-%m-%d %H:%M:%S")
                    else:
                        release_time_readable = None

                    metadata_end_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                    logging.info(color(f"{metadata_end_time}", fore="2d75ce"))
                    logging.info(color(f"Metadata extraction completed successfully.\n", fore="00eaa0"))

                except (urllib.error.URLError, socket.gaierror) as e:
                    logging.info(color(f"\nA network error occurred while extracting metadata from {url}:", fore="f26400"))
                    logging.info(e)
                    time.sleep(sleep)
                    continue

                except Exception as e:
                    logging.info(color(f"\nAn undefined error occurred while extracting metadata from {url}:", fore="f26400"))
                    logging.info(f"{e}\n")
                    invalid_urls.append(url)

                except KeyboardInterrupt:
                    logging.info("\nTermination request received. Exiting...\n")
                    sys.exit()


                try:
                    extract_title_success = False
                    video_stream = None
                    video_file = f"extracted-artifact-V_{video_id}.mp4"
                    audio_file = f"extracted-artifact-A_{video_id}.mp4"
                    multiplex_file = f"multiplex-artifact_{video_id}.mp4"

                    # Create a directory structure.
                    channel_directory = create_channel_directories(video_file, channel, url)
                    directory_path = create_geolocation_directories(video_file, location, url, channel_directory)
                    video_path = os.path.join(directory_path, video_file)
                    audio_path = os.path.join(directory_path, audio_file)
                    multiplex_path = os.path.join(directory_path, multiplex_file)

                    # Verify that video file does not exist.
                    if os.path.exists(video_path):
                        logging.info(color(f'\n"{video_file}" already exists.', fore="f26400"))
                        delete_time_start = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                        logging.info(color(f"\n{delete_time_start}", fore="2d75ce"))
                        logging.info(f'Attempting to remove the directory entry for "{video_file}" and mark its allocated clusters ready for reuse. Please wait...')
                        os.unlink(video_path)
                        time.sleep(5)
                        
                        if not os.path.exists(video_path):
                            delete_time_end = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                            logging.info(color(f"\n{delete_time_end}", fore="2d75ce"))
                            logging.info(color(f'Directory entry successfully removed for "{video_file}" and the allocated clusters are ready for reuse.\n', fore="00eaa0"))
                        
                        else:
                            delete_time_fail = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                            logging.info(color(f"\n{delete_time_fail}", fore="2d75ce"))
                            logging.info(color(f'Unable to perform the operation for "{video_file}".\n\n', fore="f26400"))
                            break


                    # Verify that audio file does not exist.
                    if os.path.exists(audio_path):
                        logging.info(color(f'"{audio_file}" already exists.', fore="f26400"))
                        delete_time_start = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                        logging.info(color(f"\n{delete_time_start}", fore="2d75ce"))
                        logging.info(f'Attempting to remove the directory entry for "{audio_file}" and mark its allocated clusters ready for reuse. Please wait...')
                        os.unlink(audio_path)
                        time.sleep(5)
                        
                        if not os.path.exists(audio_path):
                            delete_time_end = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                            logging.info(color(f"\n{delete_time_end}", fore="2d75ce"))
                            logging.info(color(f'Directory entry successfully removed for "{audio_file}" and the allocated clusters are ready for reuse.\n', fore="00eaa0"))
                        
                        else:
                            delete_time_fail = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                            logging.info(color(f"\n{delete_time_fail}", fore="2d75ce"))
                            logging.info(color(f'Unable to perform the operation for "{audio_file}\n\n".', fore="f26400"))
                            break


                    # Verify that multiplex file does not exist.
                    if os.path.exists(multiplex_path):
                        logging.info(color(f'"{multiplex_file}" already exists.', fore="f26400"))
                        delete_time_start = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                        logging.info(color(f"\n{delete_time_start}", fore="2d75ce"))
                        logging.info(f'Attempting to remove the directory entry for "{multiplex_file}" and mark its allocated clusters ready for reuse. Please wait...')
                        os.unlink(multiplex_path)
                        time.sleep(5)
                        
                        if not os.path.exists(multiplex_path):
                            delete_time_end = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                            logging.info(color(f"\n{delete_time_end}", fore="2d75ce"))
                            logging.info(color(f'Directory entry successfully removed for "{multiplex_file}" and the allocated clusters are ready for reuse.\n\n', fore="00eaa0"))
                        
                        else:
                            delete_time_fail = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                            logging.info(color(f"\n{delete_time_fail}", fore="2d75ce"))
                            logging.info(color(f'Unable to perform the operation for "{multiplex_file}".\n\n', fore="f26400"))
                            break

                except exceptions.PytubeError as e:
                    logging.info(color(f"\nAn error occurred while processing the stream:", fore="f26400"))
                    logging.info(e)
                    logging.info("Retrying in {sleep} seconds...\n")
                    time.sleep(sleep)
                    retry_count += 1
                
                except Exception as e:
                    logging.info(color(f"\nAn undefined error occurred during directory structure setup.", fore="f26400"))
                    logging.info(f"{e}\n")
                    video_stream = None
                    continue

                except KeyboardInterrupt:
                    logging.info("\nTermination request received. Exiting...\n")
                    sys.exit()


                UHD = "2160p"
                QHD = "1440p"
                FHD = "1080p"

                phase4_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                logging.info("\n-------------------------      PHASE 4      -------------------------\n")
                logging.info(color(f"\n{phase4_time}", fore="2d75ce"))

                while retry_count < max_retries or network_error < network_retry:
                    try:
                        while not video_stream:
                            try:
                                logging.info(f"Attempting to extract {UHD} stream.")
                                video_stream = yt.streams.filter(resolution=UHD).first()
                            
                                if not video_stream:
                                    logging.info(f"No {UHD} stream available. Attempting to extract {QHD} stream instead...")
                                    video_stream = yt.streams.filter(resolution=QHD).first()
                                
                                if not video_stream:
                                    logging.info(f"No {QHD} stream available. Attempting to extract {FHD} stream instead...")
                                    video_stream = yt.streams.filter(resolution=FHD).first()
                                
                                if not video_stream:
                                    logging.info(f"No {FHD} stream available. Attempting to extract the highest available stream...")
                                    video_stream = yt.streams.filter(only_video=True).get_highest_resolution()

                                if not video_stream:
                                    video_stream = yt.streams.filter(only_video=True).order_by('resolution').desc().first()
                                
                                if not video_stream:
                                    raise ValueError(color("No available stream found.\n", fore="f26400"))

                            except VideoUnavailable:
                                logging.info(color(f"\nThe stream {url} is unavailable.\n", fore="f26400"))
                                invalid_urls.append(url)
                                break

                            except exceptions.PytubeError as e:
                                logging.info(color(f"\nAn error occurred while processing the stream:", fore="f26400"))
                                logging.info(e)
                                logging.info("Retrying in {sleep} seconds...\n")
                                time.sleep(sleep)
                                retry_count += 1

                            except TypeError as e:
                                logging.info(color(f"\nAn error occurred while processing the stream:", fore="f26400"))
                                logging.info(e)
                                logging.info("Retrying in {sleep} seconds...\n")
                                time.sleep(sleep)
                                retry_count += 1
                        
                            except Exception as e:
                                logging.info(color(f"\nAn error occurred while processing {url}.", fore="f26400"))
                                logging.info(f"{e}\n")
                                break
                        
                            
                        download_start_time_V = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                        
                        logging.info(color(f"\n{download_start_time_V}", fore="2d75ce"))
                        logging.info(f"Extracting video #{number}: {video_title}\n")
                        logging.info("Please wait patiently...")
                        
                        video_stream.download(output_path=directory_path, filename=video_file)
                    
                        download_end_time_V = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                        logging.info(color(f"\n{download_end_time_V}", fore="2d75ce"))

                        if os.path.exists(video_path):
                            logging.info(color(f"Successfully completed extraction of video stream from {url}.", fore="00eaa0"))
                            logging.info(f"{video_path}\n")
                        
                        else:
                            break

                        
                        logging.info("\n-------------------------      PHASE 5      -------------------------\n")
                        signature_V_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                        logging.info(color(f"\n{signature_V_time}", fore="2d75ce"))
                        logging.info(f'Calculating file signatures for "{video_file}":\n')

                        video_blake3 = blake3(open(video_path, 'rb').read()).hexdigest()
                        video_sha512 = hashlib.sha512(open(video_path, 'rb').read()).hexdigest()
                        video_sha256 = hashlib.sha256(open(video_path, 'rb').read()).hexdigest()
                        video_sha1 = hashlib.sha1(open(video_path, 'rb').read()).hexdigest()
                        video_md5 = hashlib.md5(open(video_path, 'rb').read()).hexdigest()

                        logging.info(color(f"BLAKE (BLAKE3):", fore="f0f0f0"))
                        logging.info(f"{video_blake3}\n")
                                     
                        logging.info(color(f"Secure Hash Algorithm (SHA-512):", fore="f0f0f0"))
                        logging.info(f"{video_sha512}\n")
                                     
                        logging.info(color(f"Secure Hash Algorithm (SHA-256):", fore="f0f0f0"))
                        logging.info(f"{video_sha256}\n")
                                     
                        logging.info(color(f"Secure Hash Algorithm (SHA-1):", fore="f0f0f0"))
                        logging.info(f"{video_sha1}\n")
                                     
                        logging.info(color(f"Message Digest (MD5):", fore="f0f0f0"))
                        logging.info(f"{video_md5}\n")
                        
                        logging.info(color(f'Successfully completed file signature calculations for "{video_file}".\n', fore="00eaa0"))

                        
                        # Download audio stream
                        audio_stream = None

                        audio_initiation_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                        logging.info("\n-------------------------      PHASE 6      -------------------------\n")
                        logging.info(color(f"\n{audio_initiation_time}", fore="2d75ce"))
                        logging.info(f"Preparing to extract audio stream for {url}. Please wait...\n")

                        


                        while not audio_stream:
                            try:
                                download_start_time_A = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                                logging.info(color(f"\n{download_start_time_A}", fore="2d75ce"))
                                logging.info(f"Attempting to extract audio stream for {url}.\n")
                                audio_stream = yt.streams.filter(only_audio=True).order_by('abr').last()
                                
                                
                                logging.info(f"Extracting audio stream for video #{number}: {video_title}\n")
                                logging.info("Please wait patiently...")
                                
                                audio_stream.download(output_path=directory_path, filename=audio_file)
                                
                                download_end_time_A = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                                logging.info(color(f"{download_end_time_A}", fore="2d75ce"))

                                if os.path.exists(audio_path):
                                    logging.info(color(f"Successfully completed extraction of audio stream from {url}.", fore="00eaa0"))
                                    logging.info(f"{audio_path}\n")
                                
                                else:
                                    break
                            
                            except exceptions.PytubeError as e:
                                audio_exception_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                                logging.info(color(f"\n{audio_exception_time}", fore="2d75ce"))
                                logging.info(color("\nAn error occurred: {e}\n", fore="f26400"))
                                logging.info('Reinitiating download for "{video_file}" in {sleep} seconds. Please wait...\n')
                                time.sleep(sleep)

                        logging.info("\n-------------------------      PHASE 7      -------------------------\n")

                        signature_A_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                        logging.info(color(f"\n{signature_A_time}", fore="2d75ce"))
                        logging.info(f'Calculating file signatures for "{audio_file}":\n')

                        audio_blake3 = blake3(open(audio_path, 'rb').read()).hexdigest()
                        audio_sha512 = hashlib.sha512(open(audio_path, 'rb').read()).hexdigest()
                        audio_sha256 = hashlib.sha256(open(audio_path, 'rb').read()).hexdigest()
                        audio_sha1 = hashlib.sha1(open(audio_path, 'rb').read()).hexdigest()
                        audio_md5 = hashlib.md5(open(audio_path, 'rb').read()).hexdigest()


                        logging.info(color(f"BLAKE (BLAKE3):", fore="f0f0f0"))
                        logging.info(f"{audio_blake3}\n")
                                     
                        logging.info(color(f"Secure Hash Algorithm (SHA-512):", fore="f0f0f0"))
                        logging.info(f"{audio_sha512}\n")
                                     
                        logging.info(color(f"Secure Hash Algorithm (SHA-256):", fore="f0f0f0"))
                        logging.info(f"{audio_sha256}\n")
                                     
                        logging.info(color(f"Secure Hash Algorithm (SHA-1):", fore="f0f0f0"))
                        logging.info(f"{audio_sha1}\n")
                                     
                        logging.info(color(f"Message Digest (MD5):", fore="f0f0f0"))
                        logging.info(f"{audio_md5}\n")

                        logging.info(color(f'Successfully completed file signature calculations for "{audio_file}".\n', fore="00eaa0"))


                        # Multiplex video and audio streams into a single file
                        logging.info("\n-------------------------      PHASE 8      -------------------------\n")
                        multiplex_start = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                        logging.info(color(f"\n{multiplex_start}", fore="2d75ce"))
                        logging.info(f"Initiating multiplex procedure. Please wait...\n")

                        multiplex_exists = False

                        while not multiplex_exists:
                            os.system(f'ffmpeg -hide_banner -loglevel panic -i "{video_path}" -i "{audio_path}" -c:v copy -c:a aac -map_metadata 0 "{multiplex_path}"')
                            multiplex_end = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                            
                            if os.path.exists(multiplex_path):
                                logging.info(color(f"Successfully completed multiplex procedure.", fore="00eaa0"))
                                logging.info(f"{multiplex_path}\n")
                                multiplex_exists = True
                                break

                            else:
                                logging.info(color(f"Failed to create {multiplex_file}. Retrying in 3 seconds.", fore="f26400"))
                                time.sleep(3)
                                multiplex_exists = False

                        
                        logging.info("\n-------------------------      PHASE 9      -------------------------\n")
                        signature_M_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
                        logging.info(color(f"\n{signature_M_time}", fore="2d75ce"))
                        logging.info(f'Calculating file signature for "{multiplex_file}":\n')

                        multiplex_sha512 = hashlib.sha512(open(multiplex_path, 'rb').read()).hexdigest()

                        logging.info(color(f"Secure Hash Algorithm (SHA-512):", fore="f0f0f0"))
                        logging.info(f"{multiplex_sha512}\n")
                        
                        logging.info(color(f'Successfully completed the file signature calculation for "{multiplex_file}".\n\n\n', fore="00eaa0"))


                        logging.info("---------------------------------------------------------------------")
                        logging.info(color(f"\t\t<<<  EXTRACTION COMPLETED SUCCESSFULLY  >>>", fore="00eaa0"))
                        logging.info(f"\t\t{url}")
                        logging.info(f"---------------------------------------------------------------------\n\n\n")

                        # Append collected data to a dictionary
                        writer.writerow([video_file, video_path, url, video_title, view_count, like_count, comment_count, location, duration, channel, download_end_time_V, release_time_readable, upload_date_readable, video_blake3, video_sha512, video_sha256, video_sha1, video_md5, audio_file, audio_path, download_end_time_A, audio_blake3, audio_sha512, audio_sha256, audio_sha1, audio_md5, multiplex_file, multiplex_path, multiplex_sha512, multiplex_start, multiplex_end])
                        break


                    except http.client.IncompleteRead:
                        sleep = 10
                        retry_count += 1

                        if retry_count >= max_retries:
                            logging.info(color(f"\nUnable to download video stream from {url} after {max_retries} attempts.\n", fore="f26400"))
                            break

                        logging.info(color(f"\nAn error occurred while attempting to download video stream from {url}: Incomplete read.", fore="f26400"))
                        logging.info(f"Reinitiating download after {sleep} seconds...\n")
                        time.sleep(sleep)
                        continue

                    except (urllib.error.URLError, socket.gaierror) as e:
                        logging.info(color(f"\nA network error occurred while attempting to download video stream from {url}:", fore="f26400"))
                        logging.info(e)
                        network_error += 1
                        time.sleep(sleep)
                        continue

                    except KeyError as e:
                        logging.info(color(f"\nAn error occurred:", fore="f26400"))
                        logging.info(e)
                        time.sleep(sleep)
                        continue

                    except VideoUnavailable:
                        logging.info(color(f"\nAn error occurred: Video stream from {url} is unavailable.\n", fore="f26400"))
                        invalid_urls.append(url)
                        break

                    except (AttributeError, ValueError) as e:
                        logging.info(color(f"\nAn error occurred while attempting to download video stream from {url}:", fore="f26400"))
                        logging.info(f"{e}\n")
                        retry_count += 1
                        continue

                    except exceptions.RegexMatchError:
                        logging.info(color(f"\nInvalid URL detected:", fore="f26400"))
                        logging.info(url)
                        invalid_urls.append(url)
                        break

                    except ConnectionError as e:
                        logging.info(color("\nNetwork error detected:", fore="f26400"))
                        logging.info(e)
                        logging.info("\nPlease verify that a network connection exists.")
                        network_error += 1
                        time.sleep(sleep)

                        if network_error >= network_retry:
                            logging.info(color("\nNo network connection detected. Exiting...", fore="f26400"))
                            sys.exit(1)

                    except Exception as e:
                        logging.info(color(f"\nAn undefined error occurred:", fore="f26400"))
                        logging.info(f"{e}\n")
                        network_error += 1
                        invalid_urls.append(url)

                        if network_error >= network_retry:
                            logging.info(color("Unable to resolve error. Proceeding...\n\n", fore="f26400"))
                            break

                        else:
                            continue

                    except KeyboardInterrupt:
                        logging.info("\nTermination request received. Exiting...\n")
                        sys.exit()

        
        # Handle invalid URL entries.
        if invalid_urls:
            invalid_file = "invalid_urls.csv"
            with open(invalid_file, 'w', encoding='utf-8') as invalid:
                writer = csv.writer(invalid)
                writer.writerow(invalid_urls)

        return
    
    except PermissionError as e:
        logging.info(color(f"\nA permission error was encountered when processing {csv_log}:", fore="f26400"))
        logging.info(f"{e}\n")

    except Exception as e:
        logging.info(color(f"\nAn undefined error occurred while processing {csv_log}:", fore="f26400"))
        logging.info(f"{e}\n")



# Function for creating directory structure and moving files into them.
def create_channel_directories(video_file, channel, url):
    
    logging.info("\n-------------------------      PHASE 3      -------------------------\n")
    directory_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
    logging.info(color(f"\n{directory_time}", fore="2d75ce"))
    logging.info(f'Initiating directory structure setup for "{video_file}".\n')
    
    try:
        if channel is not None:
            sanitized_channel = re.sub(r"[^a-zA-Z0-9 äåöÄÅÖ_-]", "", channel)
            logging.info(f'Channel data identified in "{video_file}": ' + (color(f'{sanitized_channel}\n', fore="00eaa0")))
            sanitized_channel = re.sub(r"[ ]", "_", sanitized_channel)
            channel_directory = sanitized_channel
            
            if not os.path.exists(channel_directory):
                os.makedirs(channel_directory)
                logging.info(color(f'Directory successfully created: "{channel_directory}"\n', fore="00eaa0"))
            
            channel_path = os.path.abspath(channel_directory)
            return channel_directory
        
        else:
            channel_directory = "NO_CHANNEL_DATA"
            logging.info(f'No channel data identified in "{video_file}"\n.')

            if not os.path.exists(channel_directory):
                os.makedirs(channel_directory)
                logging.info(color(f'Directory successfully created: "{channel_directory}"\n', fore="00eaa0"))
            
            channel_path = os.path.abspath(channel_directory)
            return channel_directory
        

    except FileNotFoundError as e:
        logging.info(color(f"An error occurred:", fore="f26400"))
        logging.info(e)
        return None
        
    except Exception as e:
        logging.info(color("An undefined error occurred:", fore="f26400"))
        logging.info(e)
        return None
    
    except KeyboardInterrupt:
        logging.info("\nTermination request received. Exiting...\n")
        sys.exit()


def create_geolocation_directories(video_file, location, url, channel_directory):
    
    try:
        if location is not None:
            sanitized_location = re.sub(r"[^a-zA-Z0-9 äåöÄÅÖ_-]", "", location)
            logging.info(f'Geolocation data identified in "{video_file}": ' + (color(f'{sanitized_location}\n', fore="00eaa0")))
            sanitized_location = re.sub(r"[ ]", "_", sanitized_location)
            location_directory = os.path.join(channel_directory, sanitized_location)
            
            if not os.path.exists(location_directory):
                os.makedirs(location_directory)
                logging.info(color(f'Directory successfully created: "{location_directory}"\n', fore="00eaa0"))
            
            directory_path = os.path.abspath(location_directory)
            return directory_path

        else:
            location_directory = "NO_GEO_DATA"
            logging.info(f'No geolocation data identified in "{video_file}".')
            location_directory_path = os.path.join(channel_directory, location_directory)

            if not os.path.exists(location_directory_path):
                os.makedirs(location_directory_path)
                logging.info(color(f'\nDirectory successfully created: "{location_directory_path}"\n', fore="00eaa0"))
            
            directory_path = os.path.abspath(location_directory_path)
            return directory_path

    except FileNotFoundError as e:
        logging.info(color(f"An error occurred:", fore="f26400"))
        logging.info(e)
        return None
        
    except Exception as e:
        logging.info(color("An undefined error occurred:", fore="f26400"))
        logging.info(e)
        return None
    
    except KeyboardInterrupt:
        logging.info("\nTermination request received. Exiting...\n")
        sys.exit()



def initiator():
        urls = read_urls_from_csv()
        sanitized_urls = sanitize_urls(urls)
        csv_log = check_csv()
        download_data(sanitized_urls, csv_log)


try:

    # Initialize logging functionality.
    process_log = "process.log"
    logging.basicConfig(filename=process_log, level=logging.INFO)
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter('%(message)s')
    console.setFormatter(formatter)
    logging.getLogger('').addHandler(console)

except KeyboardInterrupt:
    logging.info("\nTermination request received. Exiting...\n")
    sys.exit()

except PermissionError as e:
    logging.info(color("Insufficient permissions detected:", fore="f26400"))
    logging.info(f"{e}\n")

except Exception as e:
    logging.info(color("An undefined error has occurred:", fore="f26400"))
    logging.info(f"{e}\n")


if __name__ == '__main__':
    initiator()
